package app

import (
	"context"
	"os/signal"
	"syscall"
	"time"

	"github.com/sirupsen/logrus"
	"gitea.pena/PenaSide/discount/internal/core"
	"gitea.pena/PenaSide/discount/internal/initialize"
	"gitea.pena/PenaSide/discount/internal/server"
	"gitea.pena/PenaSide/discount/pkg/closer"
	"gitea.pena/PenaSide/discount/pkg/mongo"
)

const (
	shutdownTimeout = 5 * time.Second
)

func Run(config *core.Config, logger *logrus.Logger) {
	ctx, cancel := signal.NotifyContext(context.Background(), syscall.SIGINT, syscall.SIGTERM)
	defer cancel()

	closer := closer.New()

	database, err := mongo.Connect(ctx, &mongo.ConnectDeps{
		Configuration: &config.Database.Connection,
		Timeout:       10 * time.Second,
	})
	if err != nil {
		logger.Fatalln("failed connection to db: ", err)
		panic(err)
	}

	repositories := initialize.NewRepositories(&initialize.RepositoriesDeps{
		Logger:   logger,
		Database: database,
	})

	services := initialize.NewServices(&initialize.ServicesDeps{
		Logger:       logger,
		Repositories: repositories,
	})

	controllers := initialize.NewControllers(&initialize.ControllersDeps{
		Logger:   logger,
		Services: services,
	})

	serverGRPC := server.NewGRPC(logger).Register(controllers.DiscountController)
	serverHTTP := server.NewHTTP(logger).Register(ctx, &config.GRPC)

	go serverGRPC.Run(&config.GRPC)
	go serverHTTP.Run(&config.HTTP)

	closer.Add(database.Client().Disconnect)
	closer.Add(serverGRPC.Stop)

	<-ctx.Done()

	logger.Info("shutting down app gracefully")

	shutdownCtx, cancel := context.WithTimeout(context.Background(), shutdownTimeout)

	defer cancel()

	if err := closer.Close(shutdownCtx); err != nil {
		logger.Errorf("failed to close with graceful shutdown: %v", err)
	}
}
