package controller

import (
	"context"

	"github.com/sirupsen/logrus"
	"google.golang.org/protobuf/types/known/emptypb"
	"gitea.pena/PenaSide/discount/internal/core"
	"gitea.pena/PenaSide/discount/internal/models"
	"gitea.pena/PenaSide/discount/internal/proto/discount"
	"gitea.pena/PenaSide/discount/internal/utils/transfer"
)

type DiscountService interface {
	ApplyDiscounts(context.Context, *discount.ApplyDiscountRequest) (*discount.ApplyDiscountResponse, error)
	GetDiscountByID(ctx context.Context, id string) (*models.Discount, error)
	GetDiscountsByUserID(ctx context.Context, id string) ([]models.Discount, error)
	GetAllDiscounts(ctx context.Context) ([]models.Discount, error)
	DeleteDiscount(ctx context.Context, id string) (*models.Discount, error)
	CreateDiscount(context.Context, *models.Discount) (*models.Discount, error)
	UpdateDiscount(ctx context.Context, discountID string, request *core.DiscountOptional) (*models.Discount, error)
}

type DiscountControllerDeps struct {
	Logger          *logrus.Logger
	DiscountService DiscountService
}

/*
TODO:

1) Добавить трассировку
2) Добавить валидацию
*/

type DiscountController struct {
	logger          *logrus.Logger
	discountService DiscountService

	discount.UnimplementedDiscountServiceServer
}

func NewDiscountController(deps *DiscountControllerDeps) *DiscountController {
	return &DiscountController{
		discountService: deps.DiscountService,
		logger:          deps.Logger,
	}
}

func (receiver *DiscountController) ApplyDiscounts(ctx context.Context, request *discount.ApplyDiscountRequest) (*discount.ApplyDiscountResponse, error) {
	response, err := receiver.discountService.ApplyDiscounts(ctx, request)
	if err != nil {
		return nil, determineError("failed to apply discounts: %v", err)
	}

	return response, nil
}

func (receiver *DiscountController) GetDiscountByID(ctx context.Context, request *discount.GetDiscountByIDRequest) (*discount.Discount, error) {
	findedDiscount, err := receiver.discountService.GetDiscountByID(ctx, request.ID)
	if err != nil {
		return nil, determineError("failed to get discount by id: %v", err)
	}

	return transfer.DiscountModelToProto(*findedDiscount), nil
}

func (receiver *DiscountController) CreateDiscount(ctx context.Context, request *discount.CreateDiscountRequest) (*discount.Discount, error) {
	target := transfer.DiscountCalculationTargetProtoToModel(request.Target)
	condition := transfer.DiscountConditionProtoToModel(request.Condition)

	findedDiscount, err := receiver.discountService.CreateDiscount(ctx, &models.Discount{
		Target:      *target,
		Condition:   *condition,
		Name:        request.Name,
		Description: request.Description,
		Layer:       request.Layer,
	})
	if err != nil {
		return nil, determineError("failed to create discount: %v", err)
	}

	return transfer.DiscountModelToProto(*findedDiscount), nil
}

func (receiver *DiscountController) DeleteDiscount(ctx context.Context, request *discount.GetDiscountByIDRequest) (*discount.Discount, error) {
	deletedDiscount, err := receiver.discountService.DeleteDiscount(ctx, request.ID)
	if err != nil {
		return nil, determineError("failed to delete discount: %v", err)
	}

	return transfer.DiscountModelToProto(*deletedDiscount), nil
}

func (receiver *DiscountController) GetUserDiscounts(ctx context.Context, request *discount.GetDiscountByIDRequest) (*discount.Discounts, error) {
	userDiscounts, err := receiver.discountService.GetDiscountsByUserID(ctx, request.ID)
	if err != nil {
		return nil, determineError("failed to get user discounts: %v", err)
	}

	return transfer.DiscountsModelToProto(userDiscounts), nil
}

func (receiver *DiscountController) GetAllDiscounts(ctx context.Context, _ *emptypb.Empty) (*discount.Discounts, error) {
	discounts, err := receiver.discountService.GetAllDiscounts(ctx)
	if err != nil {
		return nil, determineError("failed to delete discount: %v", err)
	}

	return transfer.DiscountsModelToProto(discounts), nil
}

func (receiver *DiscountController) UpdateDiscount(ctx context.Context, request *discount.DiscountOptional) (*discount.Discount, error) {
	discount, err := receiver.discountService.UpdateDiscount(ctx, request.ID, transfer.OptionalDiscountProtoToCore(request))
	if err != nil {
		return nil, determineError("failed to delete discount: %v", err)
	}

	return transfer.DiscountModelToProto(*discount), nil
}
