package service

import (
	"context"
	"github.com/sirupsen/logrus"
	"gitea.pena/PenaSide/discount/internal/core"
	"gitea.pena/PenaSide/discount/internal/models"
	"gitea.pena/PenaSide/discount/internal/proto/discount"
	"gitea.pena/PenaSide/discount/internal/utils"
	"gitea.pena/PenaSide/discount/internal/utils/transfer"
)

type DiscountRepository interface {
	Insert(ctx context.Context, discount *models.Discount) (string, error)
	DeleteByID(ctx context.Context, discountID string) (*models.Discount, error)
	FindByID(ctx context.Context, discountID string) (*models.Discount, error)
	FindAll(ctx context.Context) ([]models.Discount, error)
	FindByUserID(ctx context.Context, userID string) ([]models.Discount, error)
	Determine(ctx context.Context, conditions *core.DiscountConditions) ([]models.Discount, uint64, error)
	UpdateOne(ctx context.Context, update *core.UpdateDiscountSettings) (*models.Discount, error)
}

type DiscountServiceDeps struct {
	Logger             *logrus.Logger
	DiscountRepository DiscountRepository
}

type DiscountService struct {
	logger             *logrus.Logger
	discountRepository DiscountRepository
}

func NewDiscountService(deps *DiscountServiceDeps) *DiscountService {
	return &DiscountService{
		logger:             deps.Logger,
		discountRepository: deps.DiscountRepository,
	}
}

func (receiver *DiscountService) ApplyDiscounts(ctx context.Context, request *discount.ApplyDiscountRequest) (*discount.ApplyDiscountResponse, error) {
	receiver.logger.Info("AD req", request)

	// Создаем условия скидки на основе запроса
	conditions, err := utils.ConstituteConditions(request)
	if err != nil {
		receiver.logger.Errorf("failed to constitute conditions on <ApplyDiscounts> of <DiscountService>: %v", err)
		return nil, ErrInvalidInputValue
	}

	// Определяем применимые скидки и итоговую цену
	discounts, price, err := receiver.discountRepository.Determine(ctx, conditions)
	if err != nil {
		receiver.logger.Errorf("failed to determine conditions on <ApplyDiscounts> of <DiscountService>: %v", err)
		return nil, err
	}

	// Если итоговая цена равна 0, вычисляем сумму цен всех продуктов из запроса
	if price == 0 {
		for _, product := range request.Products {
			price += product.Price
		}
	}

	receiver.logger.Info("AD discounts", discounts, price)

	// Возвращаем ответ с итоговой ценой и примененными скидками
	return &discount.ApplyDiscountResponse{
		Price:            price,
		AppliedDiscounts: transfer.DiscountsModelToProto(discounts).Discounts,
	}, nil
}

func (receiver *DiscountService) GetDiscountByID(ctx context.Context, id string) (*models.Discount, error) {
	discount, err := receiver.discountRepository.FindByID(ctx, id)
	if err != nil {
		receiver.logger.Errorf("failed to get discount on <GetDiscountByID> of <DiscountService>: %v", err)
		return nil, err
	}

	return discount, nil
}

func (receiver *DiscountService) GetDiscountsByUserID(ctx context.Context, id string) ([]models.Discount, error) {
	discounts, err := receiver.discountRepository.FindByUserID(ctx, id)
	if err != nil {
		receiver.logger.Errorf("failed to get discount on <GetDiscountByUserID> of <DiscountService>: %v", err)
		return nil, err
	}

	return discounts, nil
}

func (receiver *DiscountService) GetAllDiscounts(ctx context.Context) ([]models.Discount, error) {
	discounts, err := receiver.discountRepository.FindAll(ctx)
	if err != nil {
		receiver.logger.Errorf("failed to get all discounts on <GetAllDiscounts> of <DiscountService>: %v", err)
		return nil, err
	}

	return discounts, nil
}

func (receiver *DiscountService) DeleteDiscount(ctx context.Context, id string) (*models.Discount, error) {
	deletedDiscount, err := receiver.discountRepository.DeleteByID(ctx, id)
	if err != nil {
		receiver.logger.Errorf("failed delete discount on <DeleteDiscount> of <DiscountService>: %v", err)
		return nil, err
	}

	return deletedDiscount, nil
}

func (receiver *DiscountService) CreateDiscount(ctx context.Context, discount *models.Discount) (*models.Discount, error) {
	createdDiscountID, err := receiver.discountRepository.Insert(ctx, discount)
	if err != nil {
		return nil, err
	}
	if createdDiscountID == "" {
		receiver.logger.Errorf("empty discount ID on <CreateDiscount> of <DiscountService>")
		return nil, ErrInvalidReturnValue
	}

	return &models.Discount{
		Target:      discount.Target,
		Condition:   discount.Condition,
		ID:          createdDiscountID,
		Name:        discount.Name,
		Description: discount.Description,
		Layer:       discount.Layer,
	}, nil
}

func (receiver *DiscountService) UpdateDiscount(ctx context.Context, discountID string, request *core.DiscountOptional) (*models.Discount, error) {
	updatedDiscount, err := receiver.discountRepository.UpdateOne(ctx, &core.UpdateDiscountSettings{
		ID:          discountID,
		Condition:   request.Condition,
		Name:        request.Name,
		Description: request.Description,
		Layer:       request.Layer,
		Products:    &request.Target.Products,
		Factor:      &request.Target.Factor,
		TargetScope: &request.Target.TargetScope,
		TargetGroup: &request.Target.TargetGroup,
		Overhelm:    &request.Target.Overhelm,
		Deprecated:  request.Deprecated,
	})
	if err != nil {
		receiver.logger.Errorf("failed to update discount on <UpdateDiscount> of <DiscountService>")
		return nil, err
	}

	return updatedDiscount, nil
}
