package discount_test

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"gitea.pena/PenaSide/discount/internal/core"
	"gitea.pena/PenaSide/discount/internal/fields"
	"gitea.pena/PenaSide/discount/internal/models"
	"gitea.pena/PenaSide/discount/internal/utils/expression/discount"
)

func TestGetUpdateDiscountExpression(t *testing.T) {
	name := "discount"
	description := ""
	layer := uint32(3)
	deprecated := false
	user := "testUser"
	userType := "nkvo"
	coupon := "coupon1"
	product := "p1"
	priceFrom := uint64(200)
	cartPurchasesAmount := uint64(20000)
	purchasesAmount := uint64(10000)
	usage := uint64(20)
	term := uint64(40)
	factor := 0.95
	targetScope := models.TargetSum
	targetGroup := "group"
	overhelm := true
	condition := models.DiscountCondition{
		User:                &user,
		Product:             &product,
		PriceFrom:           &priceFrom,
		Group:               &targetGroup,
		Usage:               &usage,
		UserType:            &userType,
		Coupon:              &coupon,
		CartPurchasesAmount: &cartPurchasesAmount,
		PurchasesAmount:     &purchasesAmount,
		Term:                &term,
		Period: &models.PeriodCondition{
			From: time.Unix(1674546287, 0).UTC(),
			To:   time.Unix(1674546287, 0).UTC(),
		},
	}
	products := []models.ProductTarget{
		{
			ID:       "p1",
			Factor:   0.42,
			Overhelm: false,
		},
	}

	t.Run("Формирование запроса обновления с пустыми параметрами", func(t *testing.T) {
		assert.Nil(t, discount.GetUpdateDiscountExpression(&core.UpdateDiscountSettings{}))
	})

	t.Run("Формирование запроса обновления с nil", func(t *testing.T) {
		assert.Nil(t, discount.GetUpdateDiscountExpression(nil))
	})

	t.Run("Формирование запроса обновления с заполненными параметрами", func(t *testing.T) {
		update := discount.GetUpdateDiscountExpression(&core.UpdateDiscountSettings{
			ID:          "test",
			Name:        &name,
			Description: &description,
			Layer:       &layer,
			Deprecated:  &deprecated,
			Factor:      &factor,
			TargetScope: &targetScope,
			TargetGroup: &targetGroup,
			Overhelm:    &overhelm,
			Condition:   &condition,
			Products:    &products,
		})

		expression := *update.Expression

		assert.Equal(t,
			&discount.UpdateDiscountExpression{
				ID: "test",
				Expression: &bson.M{
					"$set": bson.M{
						fields.Discount.Name:                         name,
						fields.Discount.Description:                  description,
						fields.Discount.Layer:                        layer,
						fields.Discount.Deprecated:                   deprecated,
						fields.PeriodCondition.From:                  primitive.NewDateTimeFromTime(time.Unix(1674546287, 0).UTC()),
						fields.PeriodCondition.To:                    primitive.NewDateTimeFromTime(time.Unix(1674546287, 0).UTC()),
						fields.DiscountCondition.Product:             product,
						fields.DiscountCondition.PriceFrom:           priceFrom,
						fields.DiscountCondition.PurchasesAmount:     purchasesAmount,
						fields.DiscountCondition.CartPurchasesAmount: cartPurchasesAmount,
						fields.DiscountCondition.Group:               targetGroup,
						fields.DiscountCondition.Usage:               usage,
						fields.DiscountCondition.User:                user,
						fields.DiscountCondition.UserType:            userType,
						fields.DiscountCondition.Coupon:              coupon,
						fields.DiscountCondition.Term:                term,
						fields.DiscountTarget.TargetGroup:            targetGroup,
						fields.DiscountTarget.TargetScope:            targetScope,
						fields.DiscountTarget.Factor:                 factor,
						fields.DiscountTarget.Overhelm:               overhelm,
						fields.DiscountTarget.Products:               products,
						fields.Audit.UpdatedAt:                       (expression["$set"]).(bson.M)[fields.Audit.UpdatedAt],
					},
				},
			},
			update,
		)
	})

	t.Run("Формирование запроса обновления с отсутствующими параметрами", func(t *testing.T) {
		update := discount.GetUpdateDiscountExpression(&core.UpdateDiscountSettings{
			Name:        &name,
			Description: &description,
		})

		expression := *update.Expression

		assert.Equal(t,
			&discount.UpdateDiscountExpression{
				Expression: &bson.M{
					"$set": bson.M{
						fields.Discount.Name:        name,
						fields.Discount.Description: description,
						fields.Audit.UpdatedAt:      (expression["$set"]).(bson.M)[fields.Audit.UpdatedAt],
					},
				},
			},
			update,
		)
	})
}

func TestGetUpdateDiscountsExpressions(t *testing.T) {
	name1 := "discount1"
	name2 := "discount2"
	description := ""

	t.Run("Формирование запросов обновления с пустым массивом", func(t *testing.T) {
		assert.Equal(t,
			[]discount.UpdateDiscountExpression{},
			discount.GetUpdateDiscountsExpressions([]core.UpdateDiscountSettings{}),
		)
	})

	t.Run("Формирование запроса обновления с отсутствующими и пустыми параметрами, а также с одинаковыми ID", func(t *testing.T) {
		discount.GetUpdateDiscountsExpressions([]core.UpdateDiscountSettings{
			{
				ID:          "discount1",
				Name:        &name2,
				Description: &description,
			},
			{
				ID:          "discount1",
				Name:        &name1,
				Description: &description,
			},
			{
				ID:          "discount2",
				Name:        &name2,
				Description: &description,
			},
			{},
		})
	})
}
