package transfer_test

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"google.golang.org/protobuf/types/known/timestamppb"
	"gitea.pena/PenaSide/discount/internal/models"
	"gitea.pena/PenaSide/discount/internal/proto/discount"
	"gitea.pena/PenaSide/discount/internal/utils/transfer"
)

func TestDiscountsProtoToModel(t *testing.T) {
	cartPurchasesAmount := uint64(20000)
	purchasesAmount := uint64(20000)
	userID := string("13")
	userType := string("type")
	coupon := string("coupon1")
	product := string("product1")
	term := uint64(20)

	t.Run("перевод массива скидок с proto в модель с nil", func(t *testing.T) {
		assert.Equal(t, []*models.Discount{}, transfer.DiscountsProtoToModel(nil))
	})

	t.Run("перевод массива скидок с proto в модель", func(t *testing.T) {
		assert.Equal(t,
			[]*models.Discount{
				{
					ID:          "1",
					Name:        "Скидка на объём в корзине 1",
					Layer:       8,
					Description: "",
					Condition: models.DiscountCondition{
						CartPurchasesAmount: &cartPurchasesAmount,
						PurchasesAmount:     &purchasesAmount,
						User:                &userID,
						UserType:            &userType,
						Coupon:              &coupon,
						Product:             &product,
						Term:                &term,
						Period: &models.PeriodCondition{
							From: time.Unix(1674546287, 0).UTC(),
							To:   time.Unix(0, 0).UTC(),
						},
					},
					Target: models.DiscountCalculationTarget{
						TargetScope: models.TargetSum,
						Products:    []models.ProductTarget{},
						Factor:      20,
					},
				},
			},
			transfer.DiscountsProtoToModel(&discount.Discounts{
				Discounts: []*discount.Discount{
					{
						ID:          "1",
						Name:        "Скидка на объём в корзине 1",
						Layer:       8,
						Description: "",
						Condition: &discount.DiscountCondition{
							CartPurchasesAmount: &cartPurchasesAmount,
							PurchasesAmount:     &purchasesAmount,
							User:                &userID,
							UserType:            &userType,
							Coupon:              &coupon,
							Product:             &product,
							Term:                &term,
							Period: &discount.PeriodCondition{
								From: &timestamppb.Timestamp{
									Seconds: 1674546287,
								},
								To: &timestamppb.Timestamp{},
							},
						},
						Target: &discount.DiscountCalculationTarget{
							TargetScope: discount.TargetScope_Sum.Enum(),
							Products:    []*discount.ProductTarget{},
							Factor:      20,
						},
					},
				},
			}),
		)
	})
}

func TestDiscountProtoToModel(t *testing.T) {
	cartPurchasesAmount := uint64(20000)
	deletedAtModel := time.Unix(1674546287, 0).UTC()
	deletedAtProto := timestamppb.Timestamp{Seconds: 1674546287}

	t.Run("Перевод скидки с proto в модель", func(t *testing.T) {
		assert.Nil(t, transfer.DiscountProtoToModel(nil))
	})

	t.Run("Перевод скидки с proto в модель", func(t *testing.T) {
		assert.Equal(t,
			&models.Discount{
				ID:          "1",
				Name:        "Скидка на объём в корзине 4",
				Layer:       8,
				Description: "",
				Condition: models.DiscountCondition{
					CartPurchasesAmount: &cartPurchasesAmount,
				},
				Target: models.DiscountCalculationTarget{
					TargetScope: models.TargetSum,
					Products:    []models.ProductTarget{},
					Factor:      20,
				},
				Audit: models.Audit{
					UpdatedAt: time.Unix(1674546287, 0).UTC(),
					DeletedAt: &deletedAtModel,
					Deleted:   true,
					CreatedAt: time.Unix(1674546287, 0).UTC(),
				},
			},
			transfer.DiscountProtoToModel(&discount.Discount{
				ID:          "1",
				Name:        "Скидка на объём в корзине 4",
				Layer:       8,
				Description: "",
				Condition: &discount.DiscountCondition{
					CartPurchasesAmount: &cartPurchasesAmount,
				},
				Target: &discount.DiscountCalculationTarget{
					TargetScope: discount.TargetScope_Sum.Enum(),
					Products:    []*discount.ProductTarget{},
					Factor:      20,
				},
				Audit: &discount.Audit{
					UpdatedAt: &timestamppb.Timestamp{Seconds: 1674546287},
					DeletedAt: &deletedAtProto,
					Deleted:   true,
					CreatedAt: &timestamppb.Timestamp{Seconds: 1674546287},
				},
			}),
		)
	})
}

func TestDiscountsModelToProto(t *testing.T) {
	cartPurchasesAmount := uint64(20000)
	deletedAtModel := time.Unix(1674546287, 0).UTC()
	deletedAtProto := timestamppb.Timestamp{Seconds: 1674546287}

	t.Run("Перевод массива скидок из модели в proto", func(t *testing.T) {
		assert.Equal(t,
			&discount.Discounts{Discounts: []*discount.Discount{
				{
					ID:          "1",
					Name:        "Скидка на объём в корзине 5",
					Layer:       8,
					Description: "",
					Condition: &discount.DiscountCondition{
						CartPurchasesAmount: &cartPurchasesAmount,
					},
					Target: &discount.DiscountCalculationTarget{
						TargetScope: discount.TargetScope_Sum.Enum(),
						Products:    []*discount.ProductTarget{},
						Factor:      20,
						Overhelm:    new(bool),
						TargetGroup: new(string),
					},
					Audit: &discount.Audit{
						UpdatedAt: &timestamppb.Timestamp{Seconds: 1674546287},
						DeletedAt: &deletedAtProto,
						Deleted:   true,
						CreatedAt: &timestamppb.Timestamp{Seconds: 1674546287},
					},
				},
			}},
			transfer.DiscountsModelToProto([]models.Discount{
				{
					ID:          "1",
					Name:        "Скидка на объём в корзине 5",
					Layer:       8,
					Description: "",
					Condition: models.DiscountCondition{
						CartPurchasesAmount: &cartPurchasesAmount,
					},
					Target: models.DiscountCalculationTarget{
						TargetScope: models.TargetSum,
						Products:    []models.ProductTarget{},
						Factor:      20,
					},
					Audit: models.Audit{
						UpdatedAt: time.Unix(1674546287, 0).UTC(),
						DeletedAt: &deletedAtModel,
						Deleted:   true,
						CreatedAt: time.Unix(1674546287, 0).UTC(),
					},
				},
			}),
		)
	})
}

func TestDiscountModelToProto(t *testing.T) {
	cartPurchasesAmount := uint64(20000)

	t.Run("Перевод скидки из модели в proto", func(t *testing.T) {
		assert.Equal(t,
			&discount.Discount{
				ID:          "1",
				Name:        "Скидка на объём в корзине 8",
				Layer:       8,
				Description: "",
				Condition: &discount.DiscountCondition{
					CartPurchasesAmount: &cartPurchasesAmount,
				},
				Target: &discount.DiscountCalculationTarget{
					TargetScope: discount.TargetScope_Sum.Enum(),
					Products:    []*discount.ProductTarget{},
					Factor:      20,
					Overhelm:    new(bool),
					TargetGroup: new(string),
				},
				Audit: &discount.Audit{
					CreatedAt: timestamppb.New(time.Time{}),
					UpdatedAt: timestamppb.New(time.Time{}),
				},
			},
			transfer.DiscountModelToProto(models.Discount{
				ID:          "1",
				Name:        "Скидка на объём в корзине 8",
				Layer:       8,
				Description: "",
				Condition: models.DiscountCondition{
					CartPurchasesAmount: &cartPurchasesAmount,
				},
				Target: models.DiscountCalculationTarget{
					TargetScope: models.TargetSum,
					Products:    []models.ProductTarget{},
					Factor:      20,
				},
			}),
		)
	})
}
