package integration

import (
	"context"
	"crypto/md5"
	"encoding/hex"
	"encoding/json"
	"fmt"
	formatter "github.com/antonfisher/nested-logrus-formatter"
	"github.com/golang/protobuf/ptypes/timestamp"
	"github.com/sirupsen/logrus"
	"github.com/stretchr/testify/assert"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
	"log"
	"path"
	"gitea.pena/PenaSide/discount/internal/proto/discount"
	"gitea.pena/PenaSide/discount/internal/repository"
	"gitea.pena/PenaSide/discount/internal/service"
	"runtime"
	"strconv"
	"testing"
	"time"
)

type Privilege struct {
	ID          primitive.ObjectID `bson:"_id"`
	PrivilegeID string             `bson:"privilegeId"`
	Name        string             `bson:"name"`
	ServiceKey  string             `bson:"serviceKey"`
	Description string             `bson:"description"`
	Type        string             `bson:"type"`
	Value       string             `bson:"value"`
	Price       uint64             `bson:"price"`
	Amount      uint64             `bson:"amount"`
	IsDeleted   bool               `bson:"isDeleted"`
	CreatedAt   time.Time          `bson:"createdAt"`
	UpdatedAt   time.Time          `bson:"updatedAt"`
	DeletedAt   time.Time          `bson:"deletedAt"`
}

type Tariff struct {
	ID         primitive.ObjectID `bson:"_id"`
	Name       string             `bson:"name"`
	Price      uint64             `bson:"price"`
	IsCustom   bool               `bson:"isCustom"`
	Privileges []Privilege        `bson:"privileges"`
	Deleted    bool               `bson:"isDeleted"`
	CreatedAt  time.Time          `bson:"createdAt"`
	UpdatedAt  time.Time          `bson:"updatedAt"`
	DeletedAt  *time.Time         `bson:"deletedAt"`
}

type Tariff2 struct {
	Privileges []Privilege
}

type CartVariation struct {
	Tariffs         []Tariff
	UserInformation UserInformation
}

type UserInformation struct {
	Type            string
	PurchasesAmount uint64
}

type RequestResponse struct {
	Request  *discount.ApplyDiscountRequest  `json:"request"`
	Response *discount.ApplyDiscountResponse `json:"response"`
}

func getPrivileges(ctx context.Context, url string) (map[string]struct{}, error) {
	privilegesMap := make(map[string]struct{})

	clientOptions := options.Client().ApplyURI(url)
	client, err := mongo.Connect(ctx, clientOptions)
	if err != nil {
		return nil, err
	}
	defer client.Disconnect(ctx)

	collection := client.Database("admin").Collection("privileges")

	cursor, err := collection.Find(ctx, bson.D{})
	if err != nil {
		return nil, err
	}
	defer cursor.Close(ctx)

	for cursor.Next(ctx) {
		var privilege Privilege
		err := cursor.Decode(&privilege)
		if err != nil {
			return nil, err
		}
		privilegeID := privilege.PrivilegeID
		privilegesMap[privilegeID] = struct{}{}
	}

	if err := cursor.Err(); err != nil {

		return nil, err

	}

	return privilegesMap, nil
}

func removeNonePrivilegeTariffs(ctx context.Context, privileges map[string]struct{}, url string) error {
	clientOptions := options.Client().ApplyURI(url)
	client, err := mongo.Connect(ctx, clientOptions)
	if err != nil {
		return err
	}
	defer client.Disconnect(ctx)

	collection := client.Database("admin").Collection("tariffs")

	cursor, err := collection.Find(ctx, bson.D{})
	if err != nil {
		return err
	}
	defer cursor.Close(ctx)

	for cursor.Next(ctx) {
		var tariff Tariff
		err := cursor.Decode(&tariff)
		if err != nil {
			return err
		}

		for _, privilege := range tariff.Privileges {
			if _, ok := privileges[privilege.PrivilegeID]; ok {

			} else {
				filter := bson.D{{"_id", tariff.ID}}
				_, err = collection.DeleteOne(ctx, filter)
				if err != nil {
					return err
				}
				break
			}
		}
	}

	if err := cursor.Err(); err != nil {
		return err

	}

	return nil
}

func getAllTariffs(ctx context.Context, url string) ([]Tariff, error) {
	clientOptions := options.Client().ApplyURI(url)
	client, err := mongo.Connect(ctx, clientOptions)
	if err != nil {
		return nil, err
	}
	defer client.Disconnect(ctx)

	collection := client.Database("admin").Collection("tariffs")

	cursor, err := collection.Find(ctx, bson.D{})
	if err != nil {
		return nil, err
	}
	defer cursor.Close(ctx)

	var tariffs []Tariff

	for cursor.Next(ctx) {
		var tariff Tariff
		if err := cursor.Decode(&tariff); err != nil {
			return nil, err
		}
		tariffs = append(tariffs, tariff)
	}

	if err := cursor.Err(); err != nil {
		return nil, err
	}

	return tariffs, nil
}

func generateCartVariations(tariffs [][]Tariff, userInformation []UserInformation) []CartVariation {
	var cartVariations []CartVariation
	for _, tariff := range tariffs {
		for _, userInfo := range userInformation {
			cartVariations = append(cartVariations, CartVariation{
				Tariffs:         tariff,
				UserInformation: userInfo,
			})
		}
	}

	return cartVariations
}

func ToApplyDiscount(ctx context.Context, cartVariations []*discount.ApplyDiscountRequest, svc *service.DiscountService) error {
	const maxWorkers = 10

	tasks := make(chan discount.ApplyDiscountRequest, 100)
	done := make(chan struct{})

	go func() {
		for _, variant := range cartVariations {
			tasks <- *variant
		}
		close(tasks)
	}()

	client, err := mongo.NewClient(options.Client().ApplyURI("mongodb://test:test@localhost:27023/"))
	if err != nil {
		return err
	}
	defer client.Disconnect(ctx)

	err = client.Connect(ctx)
	if err != nil {
		return err
	}

	collection := client.Database("admin").Collection("apply_test2")

	for i := 0; i < maxWorkers; i++ {
		go func() {
			for variant := range tasks {
				response, err := svc.ApplyDiscounts(ctx, &variant)
				if err != nil {
					log.Printf("ошибка при применении скидок: %v", err)
					continue
				}

				reqresp := &RequestResponse{
					Request:  &variant,
					Response: response,
				}

				_, err = collection.InsertOne(ctx, reqresp)
				if err != nil {
					log.Printf("ошибка при записи в mongodb: %v", err)
					continue
				}
			}

			done <- struct{}{}
		}()
	}

	for i := 0; i < maxWorkers; i++ {
		<-done
	}

	return nil
}

func createApplyDiscountRequest(variant CartVariation) *discount.ApplyDiscountRequest {
	userInfo := variant.UserInformation
	cartPurchasesAmount := calculateCartPurchasesAmount(variant.Tariffs)
	if cartPurchasesAmount > 2000000 {
		return nil
	}
	return &discount.ApplyDiscountRequest{
		UserInformation: &discount.UserInformation{
			ID:                  primitive.NewObjectID().Hex(),
			Type:                userInfo.Type,
			PurchasesAmount:     userInfo.PurchasesAmount,
			CartPurchasesAmount: cartPurchasesAmount,
		},
		Products: tariffsToProductInformations(variant.Tariffs),
		Date:     &timestamp.Timestamp{Seconds: time.Now().Unix()},
	}
}

// подсчет из customer
func calculateCartPurchasesAmount(tariffs []Tariff) uint64 {
	sum := uint64(0)

	for _, tariff := range tariffs {
		if tariff.Price == 0 {
			privilegesSum := uint64(0)

			for _, privilege := range tariff.Privileges {
				privilegesSum += privilege.Price * privilege.Amount
			}

			sum += privilegesSum

			continue
		}

		sum += tariff.Price
	}

	return sum
}

func tariffsToProductInformations(tarrifs []Tariff) []*discount.ProductInformation {
	productInformations := []*discount.ProductInformation{}

	for _, tariff := range tarrifs {
		for _, privilege := range tariff.Privileges {
			productInformations = append(productInformations, privilegeToProductInformation(privilege))
		}

		if tariff.Price != 0 {
			for i := range productInformations {
				productInformations[i].Price = tariff.Price
			}
		}
	}

	return productInformations
}

func privilegeToProductInformation(privilege Privilege) *discount.ProductInformation {
	return &discount.ProductInformation{
		ID:    privilege.PrivilegeID,
		Price: privilege.Price * privilege.Amount,
		Group: &privilege.ServiceKey,
		Term:  &privilege.Amount,
	}
}

func Test_ApplyDiscount(t *testing.T) {
	ctx := context.Background()
	hubAdminDbURL := "mongodb://test:test@localhost:27023/"

	//получаем текущие привилегии с хаба
	privileges, err := getPrivileges(ctx, hubAdminDbURL)
	assert.NoError(t, err, nil)
	// чистим от тех привилегий которых нет наш хаб
	err = removeNonePrivilegeTariffs(ctx, privileges, hubAdminDbURL)
	assert.NoError(t, err, nil)

	// далее выбираем все тарифы которые остались
	tariffs, err := getAllTariffs(ctx, hubAdminDbURL)
	assert.NoError(t, err, nil)

	userInformation := []UserInformation{
		{
			Type:            "nko",
			PurchasesAmount: 0,
		},
		{
			Type:            "no",
			PurchasesAmount: 0,
		},
		{
			Type:            "no",
			PurchasesAmount: 100,
		},
		{
			Type:            "no",
			PurchasesAmount: 10000,
		},
		{
			Type:            "no",
			PurchasesAmount: 100000,
		},
		{
			Type:            "no",
			PurchasesAmount: 10000000,
		},
		{
			Type:            "no",
			PurchasesAmount: 1,
		},
		{
			Type:            "no",
			PurchasesAmount: 101,
		},
		{
			Type:            "no",
			PurchasesAmount: 10001,
		},
		{
			Type:            "no",
			PurchasesAmount: 100001,
		},
		{
			Type:            "no",
			PurchasesAmount: 10000001,
		},
	}
	combo := generateAllTariffCombo(tariffs)
	cartVariations := generateCartVariations(combo, userInformation)
	logger := logrus.New()

	logger.SetReportCaller(true)
	logger.SetFormatter(&formatter.Formatter{
		TimestampFormat: "02-01-2006 15:04:05",
		HideKeys:        true,
		NoColors:        false,
		NoFieldsSpace:   true,
		CustomCallerFormatter: func(frame *runtime.Frame) string {
			return fmt.Sprintf(" (%s:%s)", path.Base(frame.File), strconv.Itoa(frame.Line))
		},
	})

	clientOptions := options.Client().ApplyURI("mongodb://test:test@localhost:27023/")
	client, err := mongo.Connect(ctx, clientOptions)
	if err != nil {
		panic(err)
	}
	defer client.Disconnect(ctx)

	collection := client.Database("admin").Collection("discounts")

	rep := repository.NewDiscountRepository(collection, logger)
	deps := service.DiscountServiceDeps{Logger: logger, DiscountRepository: rep}
	svc := service.NewDiscountService(&deps)

	reqs := cartVariationsToReqs(cartVariations)
	fmt.Println(len(reqs))
	time.Sleep(5 * time.Second)
	err = ToApplyDiscount(ctx, reqs, svc)
	assert.NoError(t, err)
}

func cartVariationsToReqs(cartVariations []CartVariation) []*discount.ApplyDiscountRequest {
	var res []*discount.ApplyDiscountRequest
	for _, variant := range cartVariations {
		req := createApplyDiscountRequest(variant)
		if req != nil {
			res = append(res, req)
		}
	}
	return res
}

func generateAllTariffCombo(arr []Tariff) [][]Tariff {
	var combinations [][]Tariff
	for k := 1; k < 9; k++ {
		combination := make([]Tariff, k)
		var inCombo func(start, depth int)
		inCombo = func(start, depth int) {
			if depth == k {
				uniquePrivileges := make(map[string]bool)
				for _, tariff := range combination {
					for _, privilege := range tariff.Privileges {
						if uniquePrivileges[privilege.PrivilegeID] {
							return
						}
						uniquePrivileges[privilege.PrivilegeID] = true
					}
				}
				combinations = append(combinations, append([]Tariff(nil), combination...))
				return
			}
			for i := start; i < len(arr); i++ {
				combination[depth] = arr[i]

				inCombo(i+1, depth+1)

			}
		}
		inCombo(0, 0)
	}

	return combinations
}

func cleanTariffData(tariffs []Tariff) []Tariff {
	tariffMap := make(map[string]Tariff)
	for _, i := range tariffs {
		var a Tariff2
		a.Privileges = i.Privileges
		privilegesBytes := []byte{}
		for _, privilege := range a.Privileges {
			privilegeBytes, err := json.Marshal(privilege)
			if err != nil {
				log.Fatal(err)
			}
			privilegesBytes = append(privilegesBytes, privilegeBytes...)
		}
		hash := md5.Sum(privilegesBytes)
		hex := hex.EncodeToString(hash[:])
		if _, ok := tariffMap[hex]; ok {
			continue
		}
		tariffMap[hex] = i
	}

	var result []Tariff
	for _, v := range tariffMap {
		result = append(result, v)
	}
	return result
}

func TestAplly2(t *testing.T) {
	logger := logrus.New()
	ctx := context.Background()
	logger.SetReportCaller(true)
	logger.SetFormatter(&formatter.Formatter{
		TimestampFormat: "02-01-2006 15:04:05",
		HideKeys:        true,
		NoColors:        false,
		NoFieldsSpace:   true,
		CustomCallerFormatter: func(frame *runtime.Frame) string {
			return fmt.Sprintf(" (%s:%s)", path.Base(frame.File), strconv.Itoa(frame.Line))
		},
	})

	clientOptions := options.Client().ApplyURI("mongodb://test:test@localhost:27023/")
	client, err := mongo.Connect(ctx, clientOptions)
	if err != nil {
		panic(err)
	}
	defer client.Disconnect(ctx)

	collection := client.Database("admin").Collection("discounts")

	rep := repository.NewDiscountRepository(collection, logger)
	deps := service.DiscountServiceDeps{Logger: logger, DiscountRepository: rep}
	svc := service.NewDiscountService(&deps)

	p := make([]*discount.ProductInformation, 2)
	group := "squiz"
	term := uint64(365)
	p[0] = &discount.ProductInformation{
		ID: "quizUnlimTime", Price: term * 3400, Term: &term, Group: &group,
	}
	group1 := "squiz"
	term1 := uint64(100)
	p[1] = &discount.ProductInformation{ID: "quizCnt", Price: term * 2000, Term: &term1, Group: &group1}

	req := &discount.ApplyDiscountRequest{
		UserInformation: &discount.UserInformation{
			ID: "65f860ed94a6ed2b2b88cf85", Type: "no", CartPurchasesAmount: term*3400 + term1*2000,
		},
		Products: p,
	}

	resp, err := svc.ApplyDiscounts(context.Background(), req)

	fmt.Println(resp)
}
